#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64


def _to_bytes(string):
    """Coerce a string into bytes

    Since Python 3 now handles bytes and str differently, this helper
    will coerce a string to bytes if possible for use with base64
    """
    try:
        string = string.encode()
    except AttributeError:
        pass
    return string


def b64encode(string):
    """Base64 encode a string to a string"""
    string = _to_bytes(string)
    return base64.b64encode(string).decode()


def b64decode(string):
    """Base64 decode a string to a string"""
    string = _to_bytes(string)
    return base64.b64decode(string).decode()



# Copyright (c) 2015 Telefonica I+D
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import subprocess

from oslo_log import log as logging

from muranoagent import bunch
import muranoagent.exceptions
from muranoagent import executors


LOG = logging.getLogger(__name__)


class ChefPuppetExecutorBase(object):
    def __init__(self, name):
        self._name = name

    def load(self, path, options):
        """Load the path and options from template into the executor.

           :param path: The path
           :param options: execution plan options.

        """
        self._path = path
        self._capture_stdout = options.get('captureStdout', True)
        self._capture_stderr = options.get('captureStderr', True)
        self._verify_exitcode = options.get('verifyExitcode', True)

    def _valid_module_name(self):
        if not self._valid_name(self._name):
            msg = ("Module recipe name format {0} is not valid".
                   format(self._name))
            LOG.debug(msg)
            raise muranoagent.exceptions.CustomException(
                0,
                message=msg,
                additional_data=None)

        self.module_name = self._name[0:self._name.rfind('::')]
        self.module_recipe = self._name[self._name.rfind('::') + 2:]

    def _valid_name(self, name):
        return '::' in name

    def _execute_command(self, command):
        stdout = subprocess.PIPE if self._capture_stdout else None
        stderr = subprocess.PIPE if self._capture_stderr else None
        process = subprocess.Popen(
            command,
            stdout=stdout,
            stderr=stderr,
            universal_newlines=True,
            cwd=os.getcwd(),
            shell=True)
        stdout, stderr = process.communicate(input)
        retcode = process.poll()

        if stdout is not None:
            if not isinstance(stdout, str):
                stdout = stdout.decode('utf-8')
            LOG.debug(u"'{0}' execution stdout: "
                      u"'{1}'".format(self.module_name, stdout))
        if stderr is not None:
            for line in stdout.splitlines():
                if 'ERROR' in line:
                    stderr += line + "\n"
            LOG.debug(u"'{0}' execution stderr: "
                      u"'{1}'".format(self.module_name, stderr))

        LOG.debug('Script {0} execution finished \
            with retcode: {1} {2}'.format(self.module_name, retcode, stderr))
        result = {
            'exitCode': retcode,
            'stdout': stdout.strip() if stdout else None,
            'stderr': stderr.strip() if stderr else None
        }
        if self._verify_exitcode and retcode != 0:
            raise muranoagent.exceptions.CustomException(
                0,
                message='Script {0} returned error code'.format(self._name),
                additional_data=result)

        return result



# Copyright (c) 2015 Telefonica I+D
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import subprocess
import yaml

from muranoagent import bunch
import muranoagent.exceptions
from muranoagent import executors
from muranoagent.executors import chef_puppet_executor_base


@executors.executor('Puppet')
class PuppetExecutor(chef_puppet_executor_base.ChefPuppetExecutorBase):

    def run(self, function, recipe_attributes=None, input=None):
        """It runs the puppet executor.

           :param function: The function
           :param recipe_attributes: recipe attributes
           :param input:
        """
        self._valid_module_name()

        try:
            self._configure_puppet()
            self._generate_files(self.module_name, self.module_recipe,
                                 recipe_attributes)
        except Exception as e:
            result = {
                'exitCode': 2,
                'stdout': None,
                'stderr': e.strerror
            }
            raise muranoagent.exceptions.CustomException(
                0,
                message='Module %s returned error code %s: %s' %
                        (self.module_name, self.module_recipe, e.strerror),
                additional_data=result)

        command = 'puppet apply --hiera_config=hiera.yaml --modulepath ' \
                  '{0} manifest.pp'.format(self._path)
        result = self._execute_command(command)
        return bunch.Bunch(result)

    def _configure_puppet(self):

        if os.path.exists('hiera.yaml'):
            return

        data = dict(
            backends='yaml',
            logger='console',
            hierarchy='%{env}',
            yaml=dict(datadir='/etc/puppet/hieradata')
        )

        self._write_yaml_file('hiera.yaml', data)

    def _generate_files(self, module, module_recipe, recipe_attributes):
        manifest = self._create_manifest(module, module_recipe)
        with open("manifest.pp", "w+") as f:
            f.write(str(manifest))

        if recipe_attributes is None:
            return

        hiera_data = self._create_hiera_data(module, recipe_attributes)
        self._write_yaml_file('default.yaml', hiera_data)

    def _create_manifest(self, module_name, module_recipe):

        if len(module_recipe) == 0:
            return "node 'default' {{ class {{ {0}:}}}}".format(module_name)
        return "node 'default' {{ class {{ {0}::{1}:}}}}".\
            format(module_name, module_recipe)

    def _create_hiera_data(self, cookbook_name,
                           recipe_attributes):
        if recipe_attributes is None:
            return
        atts = {}
        for att_name, att_value in recipe_attributes.items():
            atts[cookbook_name + '::' + att_name] = att_value

        return atts

    def _write_yaml_file(self, file, data):
        with open(file, 'w') as outfile:
            outfile.write(yaml.dump(data, default_flow_style=False))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fixtures
from oslo_config import cfg
import testtools

from muranoagent.common import config  # noqa

CONF = cfg.CONF


class MuranoAgentTestCase(testtools.TestCase):

    def setUp(self):
        super(MuranoAgentTestCase, self).setUp()
        self.useFixture(fixtures.FakeLogger('murano-agent'))

    def override_config(self, name, override, group=None):
        CONF.set_override(name, override, group)
        self.addCleanup(CONF.clear_override, name, group)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import socket
import time

import kombu
from muranoagent.common.messaging import message


class Subscription(object):
    def __init__(self, connection, queue, prefetch_count, exception_func):
        self._buffer = collections.deque()
        self._connection = connection
        self._queue = kombu.Queue(name=queue, exchange=None)
        self._consumer = kombu.Consumer(self._connection, auto_declare=False)
        self._consumer.register_callback(self._receive)
        self._consumer.qos(prefetch_count=prefetch_count)
        self._check_exception = exception_func

    def __enter__(self):
        self._consumer.add_queue(self._queue)
        self._consumer.consume()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._consumer is not None and not exc_type:
            self._consumer.cancel()
        return False

    def get_message(self, timeout=None):
        self._check_exception()
        msg_handle = self._get(timeout=timeout)
        if msg_handle is None:
            return None
        return message.Message(self._connection, msg_handle)

    def _get(self, timeout=None):
        self._check_exception()
        elapsed = 0.0
        remaining = timeout
        while True:
            time_start = time.time()
            if self._buffer:
                return self._buffer.pop()
            try:
                self._connection.drain_events(timeout=timeout and remaining)
            except socket.timeout:
                return None
            elapsed += time.time() - time_start
            remaining = timeout and timeout - elapsed or None

    def _receive(self, message_data, message):
        self._check_exception()
        self._buffer.append(message)



#    Copyright (c) 2015 Telefonica I+D
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path
from unittest import mock

from muranoagent import bunch
from muranoagent.common import config as cfg
from muranoagent import files_manager
from muranoagent.tests.unit import base

CONF = cfg.CONF


class TestFileManager(base.MuranoAgentTestCase):

    @mock.patch('os.path.isdir')
    @mock.patch('os.mkdir')
    @mock.patch('os.makedirs')
    def setUp(self, mock_makedir, mock_mkdir, mock_path):
        mock_path.return_value = True
        mock_mkdir.return_value = None
        mock_makedir.return_value = None
        super(TestFileManager, self).setUp()
        CONF.set_override('storage', 'cache')

    def test_is_svn(self):
        files = files_manager.FilesManager(self.get_template_downloable())
        self.assertTrue(files._is_svn_repository("https://sdfa/svn/ss"))

    def test_is_svn_first(self):
        files = files_manager.FilesManager(self.get_template_downloable())
        self.assertTrue(files._is_svn_repository("svn://test"))

    def test_is_svn_wrong_http_protocol(self):
        files = files_manager.FilesManager(self.get_template_downloable())
        self.assertFalse(files._is_svn_repository("httpp://sdfa/svn/ss"))

    def test_is_svn_wrong_svn_slash(self):
        files = files_manager.FilesManager(self.get_template_downloable())
        self.assertFalse(files._is_svn_repository("svn:sdfa/svn/ss"))

    @mock.patch("git.Git")
    @mock.patch('os.path.isdir')
    @mock.patch('os.makedirs')
    def test_execution_plan_type_downloable_git(self, mock_makedir, mock_path,
                                                mock_git):
        """Test an execution plan with downloadable git files

        """
        mock_makedir.return_value = None
        mock_path.return_value = True
        mock_git.clone.return_value = None
        template = self.get_template_downloable_git()
        files = files_manager.FilesManager(self.get_template_downloable())
        files._download_url_file(template.Files['mycoockbook'], "script")

    @mock.patch('os.path.isdir')
    @mock.patch('os.mkdir')
    @mock.patch('os.makedirs')
    @mock.patch('builtins.open')
    @mock.patch('requests.get')
    def test_execution_plan_type_downloable(self, mock_requests, open_mock,
                                            mock_makedir,
                                            mock_mkdir, mock_path):
        """Test an execution plan with downloadable files

        """
        mock_path.return_value = True
        mock_mkdir.return_value = None
        mock_makedir.return_value = None
        mock_requests.return_value = None
        self._open_mock(open_mock)

        template = self.get_template_downloable()
        files = files_manager.FilesManager(self.get_template_downloable())
        files._download_url_file(template.Files['file'], "script")

    @mock.patch('subprocess.Popen')
    @mock.patch('os.makedirs')
    def test_execution_plan_type_svn(self, mock_makedir, mock_subproc_popen):
        """Test an execution plan with svn files."""
        process_mock = mock.Mock()
        attrs = {'communicate.return_value': ('ouput', 'ok'),
                 'poll.return_value': 0}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        template = self.get_template_svn()
        files = files_manager.FilesManager(template)
        files._download_url_file(template.Files['file'], "script")

    @mock.patch('os.makedirs')
    def test_execution_plan_type_downloable_no_Url(self, mock_makedir):
        """It validates the URL."""
        mock_makedir.return_value = None
        template = bunch.Bunch(
            ID='ID',
            Files={
                'mycoockbook': {
                    'Name': 'mycoockbook.txt',
                    'Type': 'Downloadable'
                }
            }
        )
        files = files_manager.FilesManager(template)
        self.assertRaises(ValueError, files._download_url_file,
                          template.Files['mycoockbook'], "script")

    @mock.patch("git.Git")
    @mock.patch('os.path.isdir')
    @mock.patch('os.makedirs')
    @mock.patch('os.path.lexists')
    def test_putfile_downloable(self, mock_exists, mock_makedir,
                                path, mock_git):
        """It tests the putfile method when the file is a git URL.

        """
        path.return_value = True
        mock_git.clone.return_value = None
        mock_makedir.return_value = None
        mock_exists.return_value = True
        template = self.get_template_downloable_git()
        files = files_manager.FilesManager(template)
        for file in template.get('Files'):
            files.put_file(file, 'deploy')

    @mock.patch('builtins.open')
    @mock.patch('os.path.lexists')
    @mock.patch('os.path.isdir')
    @mock.patch('os.makedirs')
    def test_putfile_file(self, mock_makedir, mock_path,
                          mock_exists, open_mock):
        """It tests the putfile method."""
        mock_path.return_value = True
        mock_makedir.return_value = None
        mock_exists.return_value = True
        context_manager_mock = mock.Mock()
        open_mock.return_value = context_manager_mock
        file_mock = mock.Mock()
        file_mock.read.return_value = ''
        enter_mock = mock.Mock()
        enter_mock.return_value = file_mock
        exit_mock = mock.Mock()
        setattr(context_manager_mock, '__enter__', enter_mock)
        setattr(context_manager_mock, '__exit__', exit_mock)

        template = self.get_template_file()
        files = files_manager.FilesManager(template)
        for file in template.get('Files'):
            files.put_file(file, 'deploy')

    def get_template_downloable_git(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'mycoockbook': {
                    'Name': 'mycoockbook.txt',
                    'URL': 'git://github.com/tomcat.git',
                    'Type': 'Downloadable'
                }
            }
        )

    def get_template_downloable(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'file': {
                    'Name': 'myfile',
                    'URL': 'https://www.apache.org/licenses',
                    'Type': 'Downloadable'
                }
            }
        )

    def get_template_svn(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'file': {
                    'Name': 'svn',
                    'URL': 'https://mysvn/svn/repo',
                    'Type': 'Downloadable'
                }
            }
        )

    def get_template_file(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'test': {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'installer'
                }
            }
        )

    def _open_mock(self, open_mock):
        context_manager_mock = mock.Mock()
        open_mock.return_value = context_manager_mock
        file_mock = mock.Mock()
        file_mock.read.return_value = ''
        enter_mock = mock.Mock()
        enter_mock.return_value = file_mock
        exit_mock = mock.Mock()
        setattr(context_manager_mock, '__enter__', enter_mock)
        setattr(context_manager_mock, '__exit__', exit_mock)



#    Copyright (c) 2015 Telefonica I+D
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import fixtures
import ssl as ssl_module

from muranoagent import app
from muranoagent import bunch
from muranoagent.common import config as cfg
from muranoagent.common.messaging import mqclient
from muranoagent import exceptions as exc
from muranoagent.tests.unit import base
from muranoagent.tests.unit import execution_plan as ep
from muranoagent import validation

CONF = cfg.CONF


class TestApp(base.MuranoAgentTestCase, fixtures.FunctionFixture):

    @mock.patch('os.chmod')
    @mock.patch('os.path.exists')
    def setUp(self, mock_path, mock_chmod):
        super(TestApp, self).setUp()
        mock_path.side_effect = self._exists
        self.agent = app.MuranoAgent()
        CONF.set_override('storage', 'cache')
        self.addCleanup(CONF.clear_override, 'storage')

    @staticmethod
    def _exists(path):
        return 'stamp' not in path

    def test_verify_execution_plan_downloable(self):
        template = self.useFixture(ep.ExPlanDownloable()).execution_plan
        self.agent._verify_plan(template)

    def test_verify_execution_plan_wrong_format(self):
        template = bunch.Bunch(
            ID='ID',
            FormatVersion='0.0.0',
        )
        self.assertRaises(exc.IncorrectFormat,
                          validation.validate_plan, template)

    def test_verify_over_max_execution_plan(self):
        template = self.useFixture(ep.ExPlanApplication()).execution_plan
        template['FormatVersion'] = '1000.0.0'
        self.assertRaises(exc.IncorrectFormat,
                          validation.validate_plan, template)

    def test_verify_execution_application(self):
        template = self.useFixture(ep.ExPlanApplication()).execution_plan
        self.agent._verify_plan(template)

    def test_verify_wrong_execution_application(self):
        template = self.useFixture(ep.ExPlanApplication()).execution_plan
        template['Files']['ID1'] = {
            'Name': 'tomcat.git',
            'Type': 'Downloadable',
            'URL': 'https://github.com/tomcat.git'
        }
        template['FormatVersion'] = '2.0.0'
        self.assertRaises(exc.IncorrectFormat,
                          validation.validate_plan, template)

    def test_verify_execution_plan_no_files(self):
        template = self.useFixture(ep.ExPlanDownloableNoFiles()).execution_plan
        self.assertRaises(exc.IncorrectFormat,
                          validation.validate_plan, template)

    def test_verify_execution_plan_berkshelf(self):
        template = self.useFixture(ep.ExPlanBerkshelf()).execution_plan
        self.agent._verify_plan(template)

    def test_verify_execution_plan_berkshelf_wrong_version(self):
        template = self.useFixture(ep.ExPlanBerkWrongVersion()).execution_plan
        self.assertRaises(exc.IncorrectFormat,
                          validation.validate_plan, template)

    @mock.patch.object(mqclient, 'random', autospec=True)
    @mock.patch.object(mqclient, 'kombu', autospec=True)
    def test_rmq_client_initialization_with_ssl_version(self, mock_kombu,
                                                        mock_random):
        expected_heartbeat = 20  # 20 = 20 + 20 * 0, due to mocked value below.
        mock_random.random.return_value = 0

        ssl_versions = (
            ('tlsv1', getattr(ssl_module, 'PROTOCOL_TLSv1', None)),
            ('tlsv1_1', getattr(ssl_module, 'PROTOCOL_TLSv1_1', None)),
            ('tlsv1_2', getattr(ssl_module, 'PROTOCOL_TLSv1_2', None)),
            ('sslv2', getattr(ssl_module, 'PROTOCOL_SSLv2', None)),
            ('sslv23', getattr(ssl_module, 'PROTOCOL_SSLv23', None)),
            ('sslv3', getattr(ssl_module, 'PROTOCOL_SSLv3', None)))
        exception_count = 0

        for ssl_name, ssl_version in ssl_versions:
            ssl_kwargs = {
                'login': 'test_login',
                'password': 'test_password',
                'host': 'test_host',
                'port': 'test_port',
                'virtual_host': 'test_virtual_host',
                'ssl': True,
                'ssl_version': ssl_name,
                'ca_certs': ['cert1'],
                'insecure': False
            }

            # If a ssl_version is not valid, a RuntimeError is thrown.
            # According to the ssl_version docs in config.py, certain versions
            # of TLS may be available depending on the system. So, just
            # check that at least 1 ssl_version works.
            if ssl_version is None:
                e = self.assertRaises(RuntimeError, mqclient.MqClient,
                                      **ssl_kwargs)
                self.assertEqual('Invalid SSL version: %s' % ssl_name,
                                 e.__str__())
                exception_count += 1
                continue

            self.ssl_client = mqclient.MqClient(**ssl_kwargs)

            mock_kombu.Connection.assert_called_once_with(
                'amqp://{0}:{1}@{2}:{3}/{4}'.format(
                    'test_login', 'test_password', 'test_host', 'test_port',
                    'test_virtual_host'),
                heartbeat=expected_heartbeat,
                ssl={'ca_certs': ['cert1'],
                     'cert_reqs': ssl_module.CERT_REQUIRED,
                     'ssl_version': ssl_version})
            self.assertEqual(
                mock_kombu.Connection(), self.ssl_client._connection)
            self.assertIsNone(self.ssl_client._channel)
            self.assertFalse(self.ssl_client._connected)
            mock_kombu.Connection.reset_mock()

        # Check that at least one ssl_version worked.
        self.assertGreater(len(ssl_versions), exception_count)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import git
import os
import re
import requests
import shutil
import subprocess

from oslo_log import log as logging
from oslo_utils import encodeutils
import urllib

from muranoagent.common import config
from muranoagent import util

CONF = config.CONF
LOG = logging.getLogger(__name__)


class FilesManager(object):
    def __init__(self, execution_plan):
        self._fetched_files = {}
        self._files = execution_plan.get('Files') or {}

        self._cache_folder = os.path.join(
            CONF.storage, 'files', execution_plan.ID)
        if os.path.exists(self._cache_folder):
            self.clear()
        os.makedirs(self._cache_folder, 0o700)

    def put_file(self, file_id, script):
        if type(file_id) is dict:
            file_name = list(file_id.keys())[0]
            file_def = file_id[file_name]
        else:
            file_def = self._files[file_id]
            file_name = file_def['Name']

        if file_def.get('Type') == 'Downloadable':
            cache_folder = self._download_url_file(file_def, file_id)
            return self._make_symlink(cache_folder, file_name, script)
        else:
            cache_path = self._fetch_file(file_id)
            return self._make_symlink(cache_path, file_name, script)

    def _make_symlink(self, cache_path, file_name, script):
        script_folder = os.path.join(self._cache_folder, script)
        if not os.path.isdir(script_folder):
            os.mkdir(script_folder)

        file_folder = os.path.join(script_folder,
                                   os.path.dirname(file_name))
        if not os.path.isdir(file_folder):
            os.makedirs(file_folder)

        if cache_path is not None:
            script_path = os.path.join(script_folder, file_name)
            if not os.path.lexists(script_path):
                os.symlink(cache_path, script_path)
            return script_path

    def _fetch_file(self, file_id):
        if file_id in self._fetched_files:
            return self._fetched_files[file_id]

        filedef = self._files[file_id]
        out_path = os.path.join(self._cache_folder, file_id)
        body_type = filedef.get('BodyType', 'Text')
        with open(out_path, 'w') as out_file:
            if body_type == 'Text':
                out_file.write(filedef['Body'])
            elif body_type == 'Base64':
                out_file.write(util.b64decode(filedef['Body']))

        self._fetched_files[file_id] = out_path
        return out_path

    def _download_url_file(self, file_def, file_id):
        """It download the file in the murano-agent.

        It can proceed from a git file or any other internal URL
           :param file_def: file description
           :param file_id: the ID file to download
           :param input:
        """
        folder = os.path.join(self._cache_folder, file_id)
        if os.path.isdir(folder):
            return folder

        if 'URL' not in file_def:
            raise ValueError("No valid URL in file {0}".
                             format(file_def))
        url_file = file_def['URL']

        if not self._url(url_file):
            raise ValueError("Provided URL is not valid {0}".
                             format(url_file))

        if not os.path.isdir(folder):
            os.makedirs(folder)

        try:
            if self._is_git_repository(url_file):
                git.Git().clone(url_file, folder)
            elif self._is_svn_repository(url_file):
                self._download_svn(url_file, folder)
            else:
                self._download_file(url_file, folder)
        except Exception as e:
            if self._is_git_repository(url_file):
                mns = ("Error to clone the git repository {0}: {1}".
                       format(url_file, e.message))
            else:
                mns = ("Error to download the file {0}: {1}".
                       format(url_file, e.message))
            LOG.warning(mns)
            raise ValueError(mns)
        return folder

    def clear(self):
        shutil.rmtree(self._cache_folder, ignore_errors=True)

    def _download_file(self, url, path):
        local_filename = url.split('/')[-1]
        r = requests.get(url, stream=True)
        with open(os.path.join(path, local_filename), 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
                    f.flush()
        return local_filename

    def _url(self, file):
        return (urllib.parse.urlsplit(file).scheme or
                urllib.parse.urlsplit(file).netloc)

    def _is_git_repository(self, url):
        return (url.startswith(("git://",
                               "git+http://", "git+https:/"))
                or url.endswith('.git'))

    def _is_svn_repository(self, url):
        http_regex = "https?://(.*)/svn/(.*)"
        http_matches = re.search(http_regex, url)
        svn_regex = "svn://(.*)"
        svn_matches = re.search(svn_regex, url)
        if http_matches is None and svn_matches is None:
            return False
        else:
            return True

    def _download_svn(self, url_file, folder):
        self._execute_command("svn checkout {0} --non-interactive "
                              "--trust-server-cert {1}".
                              format(url_file, folder))

    def _execute_command(self, command):

        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=os.getcwd(),
            shell=True)
        stdout, stderr = process.communicate(input)
        retcode = process.poll()

        if stdout is not None:
            stdout = encodeutils.safe_decode('utf-8')
            LOG.debug(stdout)

        if stderr is not None:
            stderr = encodeutils.safe_decode('utf-8')
            LOG.error(stderr)

        if retcode != 0:
            raise ValueError(stderr)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil
import time

from cryptography.hazmat import backends
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from oslo_log import log as logging

from muranoagent import bunch
from muranoagent.common import config
from muranoagent import util

CONF = config.CONF
LOG = logging.getLogger(__name__)


class ExecutionPlanQueue(object):
    plan_filename = 'plan.json'
    result_filename = 'result.json'
    stamp_filename = 'stamp'

    def __init__(self):
        self._plans_folder = os.path.join(CONF.storage, 'plans')
        if not os.path.exists(self._plans_folder):
            os.makedirs(self._plans_folder, 0o700)
        else:
            try:
                os.chmod(self._plans_folder, 0o700)
            except OSError:
                pass
        self._key = None if not CONF.engine_key \
            else serialization.load_pem_public_key(
                CONF.engine_key, backends.default_backend())
        self._load_stamp()

    def put_execution_plan(self, execution_plan, signature, msg_id, reply_to):
        timestamp = str(int(time.time() * 10000))
        # execution_plan['_timestamp'] = timestamp
        folder_path = os.path.join(self._plans_folder, timestamp)
        os.mkdir(folder_path)
        plan_file_path = os.path.join(
            folder_path, ExecutionPlanQueue.plan_filename)
        json_plan = json.dumps({
            'Data': util.b64encode(execution_plan),
            'Signature': util.b64encode(signature or ''),
            'ID': msg_id,
            'ReplyTo': reply_to
        })
        with open(plan_file_path, 'wb') as out_file:
            out_file.write(json_plan)

    def _get_first_timestamp(self, filename):
        def predicate(folder):
            path = os.path.join(self._plans_folder, folder, filename)
            return os.path.exists(path)

        timestamps = [
            name for name in os.listdir(self._plans_folder)
            if predicate(name)
        ]
        timestamps.sort()
        return None if len(timestamps) == 0 else timestamps[0]

    def _get_first_file(self, filename):
        timestamp = self._get_first_timestamp(filename)
        if not timestamp:
            return None, None
        path = os.path.join(self._plans_folder, timestamp, filename)
        with open(path) as json_file:
            return json.loads(json_file.read()), timestamp

    def get_execution_plan(self):
        while True:
            ep_info, timestamp = self._get_first_file(
                ExecutionPlanQueue.plan_filename)
            if ep_info is None:
                return None

            try:
                data = util.b64decode(ep_info['Data'])
                if self._key:
                    signature = util.b64decode(ep_info['Signature'])
                    self._verify_signature(data, signature)

                ep = json.loads(data)
                if not isinstance(ep, dict):
                    raise ValueError('Message is not a document')

                stamp = ep.get('Stamp', -1)
                if stamp >= 0:
                    if stamp <= self._last_stamp:
                        raise ValueError('Dropping old/duplicate message')
                    self._save_stamp(stamp)

                if 'ID' not in ep:
                    ep['ID'] = ep_info['ID']
                if 'ReplyTo' not in ep:
                    ep['ReplyTo'] = ep_info['ReplyTo']

                ep['_timestamp'] = timestamp
                return bunch.Bunch(ep)
            except Exception as ex:
                LOG.exception(ex)
                self.remove(timestamp)

    def _verify_signature(self, data, signature):
        if not signature:
            raise ValueError("Required signature was not found")
        self._key.verify(
            signature,
            CONF.rabbitmq.input_queue + data,
            padding.PKCS1v15(), hashes.SHA256())

    def put_execution_result(self, result, execution_plan):
        timestamp = execution_plan['_timestamp']
        if 'ReplyTo' in execution_plan:
            result['ReplyTo'] = execution_plan.get('ReplyTo')
        path = os.path.join(
            self._plans_folder, timestamp,
            ExecutionPlanQueue.result_filename)
        with open(path, 'w') as out_file:
            out_file.write(json.dumps(result))

    def remove(self, timestamp):
        path = os.path.join(self._plans_folder, timestamp)
        shutil.rmtree(path)

    def get_execution_plan_result(self):
        return self._get_first_file(
            ExecutionPlanQueue.result_filename)

    def _load_stamp(self):
        plan_file_path = os.path.join(
            self._plans_folder, ExecutionPlanQueue.stamp_filename)
        if os.path.exists(plan_file_path):
            with open(plan_file_path) as f:
                self._last_stamp = int(f.read())
        else:
            self._last_stamp = 0

    def _save_stamp(self, stamp):
        plan_file_path = os.path.join(
            self._plans_folder, ExecutionPlanQueue.stamp_filename)
        with open(plan_file_path, 'w') as f:
            f.write(str(stamp))
            self._last_stamp = stamp



# Copyright (c) 2015 Telefonica I+D
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import subprocess

from oslo_log import log as logging

from muranoagent import bunch
import muranoagent.exceptions
from muranoagent import executors
from muranoagent.executors import chef_puppet_executor_base


LOG = logging.getLogger(__name__)


@executors.executor('Chef')
class ChefExecutor(chef_puppet_executor_base.ChefPuppetExecutorBase):

    def load(self, path, options):
        super(ChefExecutor, self).load(path, options)
        self._use_berkshelf = options.get('useBerkshelf', False)
        self._berksfile_path = options.get('berksfilePath', None)

    def run(self, function, recipe_attributes=None, input=None):
        """It runs the chef executor.

           :param function: The function
           :param recipe_attributes: recipe attributes
           :param input:
        """
        self._valid_module_name()

        cookbook_path = self._create_cookbook_path(self.module_name)

        try:
            self._configure_chef(cookbook_path)
            self._generate_manifest(self.module_name,
                                    self.module_recipe, recipe_attributes)
        except Exception as e:
            result = {
                'exitCode': 2,
                'stdout': None,
                'stderr': e.strerror
            }
            raise muranoagent.exceptions.CustomException(
                0,
                message='Cookbook {0} returned error code {1}: {2}'.format(
                    self.module_name, self.module_recipe, e.strerror,
                ), additional_data=result)

        solo_file = os.path.join(self._path, "solo.rb")
        command = 'chef-solo -j node.json -c {0}'.format(solo_file)
        result = self._execute_command(command)
        return bunch.Bunch(result)

    def _create_cookbook_path(self, cookbook_name):
        """It defines a path where all required cookbooks are located."""
        path = os.path.abspath(self._path)

        if self._use_berkshelf:
            LOG.debug('Using Berkshelf')

            # Get Berksfile
            if self._berksfile_path is None:
                self._berksfile_path = cookbook_name + '/Berksfile'
            berksfile = os.path.join(path, self._berksfile_path)
            if not os.path.isfile(berksfile):
                msg = "Berskfile {0} not found".format(berksfile)
                LOG.debug(msg)
                raise muranoagent.exceptions.CustomException(
                    0,
                    message=msg,
                    additional_data=None)

            # Create cookbooks path
            cookbook_path = os.path.join(path, "berks-cookbooks")
            if not os.path.isdir(cookbook_path):
                os.makedirs(cookbook_path)

            # Vendor cookbook and its dependencies to cookbook_path
            command = 'berks vendor --berksfile={0} {1}'.format(
                berksfile,
                cookbook_path)
            result = self._execute_command(command)
            if result['exitCode'] != 0:
                raise muranoagent.exceptions.CustomException(
                    0,
                    message='Berks returned error code',
                    additional_data=result)

            return cookbook_path

        else:
            return path

    def _configure_chef(self, cookbook_path):
        """It generates the chef files for configuration."""
        solo_file = os.path.join(self._path, 'solo.rb')
        if not os.path.exists(solo_file):
            if not os.path.isdir(self._path):
                os.makedirs(self._path)
            with open(solo_file, "w+") as f:
                f.write('cookbook_path \"' + cookbook_path + '\"')

    def _generate_manifest(self, cookbook_name,
                           cookbook_recipe, recipe_attributes):
        """It generates the chef manifest."""
        node = self._create_manifest(cookbook_name, cookbook_recipe,
                                     recipe_attributes)
        with open("node.json", "w+") as f:
            f.write(node)

    def _create_manifest(self, cookbook_name, cookbook_recipe,
                         recipe_attributes):
        node = {}
        node["run_list"] = [u"recipe[{0}::{1}]".format(
            cookbook_name, cookbook_recipe)]

        if recipe_attributes:
            node[cookbook_name] = recipe_attributes.copy()

        return json.dumps(node)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import signal
import stat
import subprocess
import sys

from oslo_log import log as logging

from muranoagent import bunch
import muranoagent.exceptions
from muranoagent import executors

LOG = logging.getLogger(__name__)


@executors.executor('Application')
class ApplicationExecutor(object):
    def __init__(self, name):
        self._name = name

    def load(self, path, options):
        self._path = path
        self._capture_stdout = options.get('captureStdout', True)
        self._capture_stderr = options.get('captureStderr', True)
        self._verify_exitcode = options.get('verifyExitcode', True)

    def run(self, function, commandline='', input=None):
        dir_name = os.path.dirname(self._path)
        os.chdir(dir_name)
        app = '"{0}" {1}'.format(os.path.basename(self._path), commandline)

        if not sys.platform == 'win32':
            os.chmod(self._path, stat.S_IEXEC | stat.S_IREAD)
            app = './' + app

        stdout = subprocess.PIPE if self._capture_stdout else None
        stderr = subprocess.PIPE if self._capture_stderr else None
        script_name = os.path.relpath(self._path)
        LOG.debug("Starting '{0}' script execution".format(script_name))

        popen_kwargs = {
            'stdout': stdout,
            'stderr': stderr,
            'universal_newlines': True,
            'cwd': dir_name,
            'shell': True
        }
        if os.name != 'nt':
            popen_kwargs['preexec_fn'] = lambda: signal.signal(
                signal.SIGPIPE, signal.SIG_DFL)

        process = subprocess.Popen(app, **popen_kwargs)
        stdout, stderr = process.communicate(input)
        retcode = process.poll()
        LOG.debug("Script {0} execution finished "
                  "with retcode: {1}".format(script_name, retcode))
        if stdout is not None:
            if hasattr(stdout, 'decode'):
                stdout = stdout.decode('utf-8')
            LOG.debug(u"'{0}' execution stdout: "
                      u"'{1}'".format(script_name, stdout))
        if stderr is not None:
            if hasattr(stderr, 'decode'):
                stderr = stderr.decode('utf-8')
            LOG.debug(u"'{0}' execution stderr: "
                      u"'{1}'".format(script_name, stderr))
        result = {
            'exitCode': retcode,
            'stdout': stdout.strip() if stdout else None,
            'stderr': stderr.strip() if stderr else None
        }
        if self._verify_exitcode and retcode != 0:
            raise muranoagent.exceptions.CustomException(
                0,
                message='Script {0} returned error code'.format(self._name),
                additional_data=result)

        return bunch.Bunch(result)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import ssl as ssl_module

import anyjson
import eventlet
import kombu
from oslo_service import sslutils

from muranoagent.common.messaging import subscription


class MqClient(object):
    def __init__(self, login, password, host, port, virtual_host,
                 ssl=False, ssl_version=None, ca_certs=None, insecure=False):
        ssl_params = None

        if ssl:
            cert_reqs = ssl_module.CERT_REQUIRED
            if insecure:
                if ca_certs:
                    cert_reqs = ssl_module.CERT_OPTIONAL
                else:
                    cert_reqs = ssl_module.CERT_NONE

            ssl_params = {
                'ca_certs': ca_certs,
                'cert_reqs': cert_reqs
            }

            if ssl_version:
                key = ssl_version.lower()
                try:
                    ssl_params['ssl_version'] = sslutils._SSL_PROTOCOLS[key]
                except KeyError:
                    raise RuntimeError("Invalid SSL version: %s" % ssl_version)

        # Time interval after which RabbitMQ will disconnect client if no
        # heartbeats were received. Usually client sends 2 heartbeats during
        # this interval. Using random to make it less lucky that many agents
        # ping RabbitMQ simultaneously
        heartbeat_rate = 20 + 20 * random.random()

        self._connection = kombu.Connection(
            'amqp://{0}:{1}@{2}:{3}/{4}'.format(
                login,
                password,
                host,
                port,
                virtual_host
            ), ssl=ssl_params, heartbeat=heartbeat_rate
        )
        self._channel = None
        self._connected = False
        self._exception = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self._connected = False
        else:
            self.close()
        return False

    def connect(self):
        self._connection.connect()
        self._channel = self._connection.channel()
        if not self._connected:
            self._connected = True
            eventlet.spawn(self._heartbeater)

    def close(self):
        if self._connected:
            self._connection.close()
            self._connected = False

    def _check_exception(self):
        ex = self._exception
        if ex:
            self._exception = None
            raise ex

    def _heartbeater(self):
        while self._connected:
            eventlet.sleep(1)
            try:
                self._connection.heartbeat_check()
            except Exception as ex:
                self._exception = ex
                self._connected = False

    def declare(self, queue, exchange='', enable_ha=False, ttl=0):
        self._check_exception()
        if not self._connected:
            raise RuntimeError('Not connected to RabbitMQ')

        queue_arguments = {}
        if enable_ha is True:
            # To use mirrored queues feature in RabbitMQ 2.x
            # we need to declare this policy on the queue itself.
            #
            # Warning: this option has no effect on RabbitMQ 3.X,
            # to enable mirrored queues feature in RabbitMQ 3.X, please
            # configure RabbitMQ.
            queue_arguments['x-ha-policy'] = 'all'
        if ttl > 0:
            queue_arguments['x-expires'] = ttl

        exchange = kombu.Exchange(exchange, type='direct', durable=True)
        queue = kombu.Queue(queue, exchange, queue, durable=False,
                            queue_arguments=queue_arguments)
        bound_queue = queue(self._connection)
        bound_queue.declare()

    def send(self, message, key, exchange=''):
        self._check_exception()
        if not self._connected:
            raise RuntimeError('Not connected to RabbitMQ')

        producer = kombu.Producer(self._connection)
        producer.publish(
            exchange=str(exchange),
            routing_key=str(key),
            body=anyjson.dumps(message.body),
            message_id=str(message.id)
        )

    def open(self, queue, prefetch_count=1):
        self._check_exception()
        if not self._connected:
            raise RuntimeError('Not connected to RabbitMQ')

        return subscription.Subscription(self._connection, queue,
                                         prefetch_count, self._check_exception)



#    Copyright (c) 2015 Telefonica I+D
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import builtins
import fixtures

from muranoagent import execution_plan_queue

from muranoagent import app
from muranoagent import bunch
from muranoagent.common import config as cfg
from muranoagent.common.messaging import mqclient
from muranoagent import exceptions as exc
from muranoagent.tests.unit import base
from muranoagent.tests.unit import execution_plan as ep
from muranoagent import validation

CONF = cfg.CONF


class TestExecutionPlanQueue(base.MuranoAgentTestCase,
                             fixtures.FunctionFixture):

    @mock.patch('os.chmod')
    @mock.patch('os.path.exists')
    def setUp(self, mock_path, mock_chmod):
        super(TestExecutionPlanQueue, self).setUp()
        mock_path.side_effect = self._exists
        self.epq = execution_plan_queue.ExecutionPlanQueue()
        CONF.set_override('storage', 'cache')
        self.addCleanup(CONF.clear_override, 'storage')

    @staticmethod
    def _exists(path):
        return 'stamp' not in path

    @mock.patch('os.path.lexists')
    @mock.patch('os.path.isdir')
    @mock.patch('os.mkdir')
    def test_put_execution_plan(self, mock_makedir, mock_path,
                                mock_exists):
        mock_path.return_value = True
        mock_makedir.return_value = None
        mock_exists.return_value = True
        mock_write = mock.mock_open()

        execution_plan = 'myplan'
        signature = None
        msg_id = 1
        reply_to = 'test'
        expected_content = ('{"Data": "bXlwbGFu", "Signature": "", '
                            '"ID": 1, "ReplyTo": "test"}')
        with mock.patch.object(builtins, 'open', mock_write) as mocked_file:
            self.epq.put_execution_plan(execution_plan, signature,
                                        msg_id, reply_to)
            mocked_file().write.assert_called_once_with(expected_content)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import time

from oslo_log import log as logging
from oslo_service import service
from oslo_utils import strutils

from muranoagent.common import config
from muranoagent.common import messaging
from muranoagent import execution_plan_queue
from muranoagent import execution_plan_runner
from muranoagent import execution_result as ex_result
from muranoagent import validation

CONF = config.CONF

LOG = logging.getLogger(__name__)


class MuranoAgent(service.Service):
    def __init__(self):
        self._queue = execution_plan_queue.ExecutionPlanQueue()
        super(MuranoAgent, self).__init__()

    @staticmethod
    def _load_package(name):
        try:
            LOG.debug('Loading plugin %s', name)
            __import__(name)
        except Exception:
            LOG.warning('Cannot load package %s', name, exc_info=True)
            pass

    def _load(self):
        path = os.path.join(os.path.dirname(__file__), 'executors')
        sys.path.insert(1, path)
        for entry in os.listdir(path):
            package_path = os.path.join(path, entry)
            if os.path.isdir(package_path):
                MuranoAgent._load_package(entry)

    def start(self):
        self._load()
        msg_iterator = self._wait_plan()
        while True:
            try:
                self._loop_func(msg_iterator)
            except Exception as ex:
                LOG.exception(ex)
                time.sleep(5)

    def _loop_func(self, msg_iterator):
        result, timestamp = self._queue.get_execution_plan_result()
        if result is not None:
            if self._send_result(result):
                self._queue.remove(timestamp)
            return

        plan = self._queue.get_execution_plan()
        if plan is not None:
            LOG.debug("Got an execution plan '{0}':".format(
                strutils.mask_password(str(plan))))
            if self._verify_plan(plan):
                self._run(plan)
            return

        next(msg_iterator)

    def _verify_plan(self, plan):
        try:
            validation.validate_plan(plan)
            return True
        except Exception as err:
            try:
                execution_result = ex_result.ExecutionResult.from_error(
                    err, plan)
                if 'ReplyTo' in plan and CONF.enable_dynamic_result_queue:
                    execution_result['ReplyTo'] = plan.ReplyTo

                self._send_result(execution_result)
            except ValueError:
                LOG.warning('Execution result is not produced')
            finally:
                return False

    def _run(self, plan):
        try:
            with execution_plan_runner.ExecutionPlanRunner(plan) as runner:
                result = runner.run()
                execution_result = ex_result.ExecutionResult.from_result(
                    result, plan)
                self._queue.put_execution_result(execution_result, plan)
        except Exception as ex:
            LOG.exception('Error running execution plan')
            execution_result = ex_result.ExecutionResult.from_error(ex,
                                                                    plan)
            self._queue.put_execution_result(execution_result, plan)

    def _send_result(self, result):
        with self._create_rmq_client() as mq:
            msg = messaging.Message()
            msg.body = result
            msg.id = result.get('SourceID')
            routing_key = CONF.rabbitmq.result_routing_key
            if ('ReplyTo' in result) and CONF.enable_dynamic_result_queue:
                routing_key = result.pop('ReplyTo')
            mq.send(message=msg,
                    key=routing_key,
                    exchange=CONF.rabbitmq.result_exchange)
        return True

    def _create_rmq_client(self):
        rabbitmq = CONF.rabbitmq
        connection_params = {
            'login': rabbitmq.login,
            'password': rabbitmq.password,
            'host': rabbitmq.host,
            'port': rabbitmq.port,
            'virtual_host': rabbitmq.virtual_host,
            'ssl': rabbitmq.ssl,
            'ssl_version': rabbitmq.ssl_version,
            'ca_certs': rabbitmq.ca_certs.strip() or None,
            'insecure': rabbitmq.insecure
        }
        return messaging.MqClient(**connection_params)

    def _wait_plan(self):
        delay = 5
        while True:
            try:
                with self._create_rmq_client() as mq:
                    with mq.open(CONF.rabbitmq.input_queue,
                                 prefetch_count=1) as subscription:
                        while True:
                            msg = subscription.get_message(timeout=5)
                            if msg is not None:
                                try:
                                    self._queue.put_execution_plan(
                                        msg.body,
                                        msg.signature,
                                        msg.id,
                                        msg.reply_to)
                                finally:
                                    msg.ack()

                            delay = 5
                            if msg is not None:
                                yield
            except KeyboardInterrupt:
                break
            except Exception:
                LOG.warning('Communication error', exc_info=True)
                time.sleep(delay)
                delay = min(delay * 1.2, 60)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ExecutorsRepo(object):
    def __init__(self):
        self._executors = {}

    def register_executor(self, name, cls):
        self._executors[name] = cls

    def create_executor(self, type, name):
        if type not in self._executors:
            return None
        return self._executors[type](name)


Executors = ExecutorsRepo()


def executor(name):
    def wrapper(cls):
        Executors.register_executor(name, cls)
        return cls
    return wrapper



#    Copyright (c) 2015 Telefonica I+D
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import fixtures

from muranoagent import bunch
from muranoagent import exceptions as ex
from muranoagent.executors import puppet
from muranoagent.tests.unit import base
from muranoagent.tests.unit import execution_plan as ep


class TestPuppetExecutor(base.MuranoAgentTestCase, fixtures.TestWithFixtures):
    def setUp(self):
        super(TestPuppetExecutor, self).setUp()
        self.puppet_executor = puppet.PuppetExecutor('module::recipe')

    def test_create_manifest(self):
        node = self.puppet_executor._create_manifest('cookbook', 'recipe')
        self.assertEqual(node, self.get_manifest('cookbook', 'recipe'))

    def test_create_manifest_norecipe(self):
        node = self.puppet_executor._create_manifest('cookbook', '')
        self.assertEqual(node, self.get_manifest_norecipe('cookbook'))

    def test_create_hierdata(self):
        atts = {
            'att1': 'value1',
            'att2': 'value2'
        }

        node = self.puppet_executor._create_hiera_data('cookbook', atts)
        self.assertEqual(node, self.get_hieradata())

    @mock.patch('builtins.open')
    def test_generate_files(self, open_mock):
        self._open_mock(open_mock)
        atts = {
            'att1': 'value1',
            'att2': 'value2'
        }

        self.puppet_executor._generate_files('cookbook', 'recipe', atts)

    @mock.patch('builtins.open')
    def test_configure_puppet(self, open_mock):
        self._open_mock(open_mock)
        self.puppet_executor._configure_puppet()

    @mock.patch('subprocess.Popen')
    @mock.patch('builtins.open')
    def test_module(self, open_mock, mock_subproc_popen):
        #
        # setup
        #
        self._open_mock(open_mock)

        process_mock = mock.Mock()
        attrs = {'communicate.return_value': ('ouput', 'ok'),
                 'poll.return_value': 0}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        template = self.useFixture(ep.PuppetExPlanDownloable()).execution_plan
        script = list(template['Scripts'].values())[0]
        self.puppet_executor.load('path',
                                  script['Options'])
        self.puppet_executor.run('test')

    @mock.patch('subprocess.Popen')
    @mock.patch('builtins.open')
    def test_module_error(self, open_mock, mock_subproc_popen):
        #
        # setup
        #
        self._open_mock(open_mock)

        process_mock = mock.Mock()
        attrs = {'communicate.return_value': ('ouput', 'error'),
                 'poll.return_value': 2}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        template = self.useFixture(ep.PuppetExPlanDownloable()).execution_plan
        script = list(template['Scripts'].values())[0]
        self.puppet_executor.load('path',
                                  script['Options'])
        self.assertRaises(ex.CustomException, self.puppet_executor.run,
                          'test')

    def test_puppet_module_wrong(self):
        puppet_executor = puppet.PuppetExecutor('wrong')
        self.assertRaises(ex.CustomException, puppet_executor.run,
                          'test')

    def _stub_uuid(self, values=None):
        class FakeUUID(object):
            def __init__(self, v):
                self.hex = v

        if values is None:
            values = []
        mock_uuid4 = mock.patch('uuid.uuid4').start()
        mock_uuid4.side_effect = [FakeUUID(v) for v in values]
        return mock_uuid4

    def _open_mock(self, open_mock):
        context_manager_mock = mock.Mock()
        open_mock.return_value = context_manager_mock
        file_mock = mock.Mock()
        file_mock.read.return_value = ''
        enter_mock = mock.Mock()
        enter_mock.return_value = file_mock
        exit_mock = mock.Mock()
        setattr(context_manager_mock, '__enter__', enter_mock)
        setattr(context_manager_mock, '__exit__', exit_mock)

    def get_hieradata(self):
        return {'cookbook::att1': 'value1', 'cookbook::att2': 'value2'}

    def get_manifest(self, cookbook, recipe):
        return "node \'default\' { " \
               "class { " + cookbook + '::' + recipe + ':}}'

    def get_manifest_norecipe(self, cookbook):
        return "node \'default\' { " \
               "class { " + cookbook + ':}}'



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class Bunch(dict):
    """Bunch dict/object-like container.

    Bunch container provides both dictionary-like and
    object-like attribute access.
    """
    def __getattr__(self, item):
        return self.__getitem__(item)

    def __setattr__(self, key, value):
        return self.__setitem__(key, value)

    def __delattr__(self, key):
        del self[key]



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

LOG = logging.getLogger("murano-common.messaging")


class Message(object):
    def __init__(self, connection=None, message_handle=None):
        self._body = None
        self._connection = connection
        self._message_handle = message_handle
        if message_handle:
            self.id = message_handle.properties.get('message_id')
            self._reply_to = message_handle.properties.get('reply_to')
            self._signature = message_handle.headers.get('signature')
        else:
            self.id = None
            self._reply_to = None
            self._signature = None

        if message_handle:
            self.body = message_handle.body

        else:
            self.body = None

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, value):
        self._body = value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value or ''

    @property
    def reply_to(self):
        return self._reply_to

    def ack(self):
        self._message_handle.ack()

    @property
    def signature(self):
        return self._signature



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa: E402

import os
import sys


# If ../muranoagent/__init__.py exists, add ../ to Python search path, so
# it will override what happens to be installed in /usr/(local/)lib/python...
possible_topdir = os.path.normpath(os.path.join(os.path.abspath(__file__),
                                                os.pardir,
                                                os.pardir,
                                                os.pardir))
if os.path.exists(os.path.join(possible_topdir,
                               'muranoagent',
                               '__init__.py')):
    sys.path.insert(0, possible_topdir)

import eventlet
eventlet.monkey_patch()
# Monkey patch the original current_thread to use the up-to-date _active
# global variable. See https://bugs.launchpad.net/bugs/1863021 and
# https://github.com/eventlet/eventlet/issues/592
import __original_module_threading as orig_threading
import threading  # noqa
orig_threading.current_thread.__globals__['_active'] = threading._active


from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service

from muranoagent import app
from muranoagent.common import config

CONF = cfg.CONF


def main():
    try:
        config.parse_args()
        logging.setup(CONF, 'muranoagent')
        launcher = service.ServiceLauncher(CONF)
        launcher.launch_service(app.MuranoAgent())
        launcher.wait()
    except RuntimeError as e:
        sys.stderr.write("ERROR: %s\n" % e)
        sys.exit(1)


if __name__ == '__main__':
    main()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import os
    import win32file

    def symlink(source, link_name):
        src = os.path.abspath(source)
        dest = os.path.abspath(link_name)
        win32file.CreateSymbolicLink(dest, src, 0)

    os.symlink = symlink
except ImportError:
    pass



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from muranoagent import bunch
from muranoagent import files_manager as fm
from muranoagent import script_runner


class ExecutionPlanRunner(object):
    def __init__(self, execution_plan):
        self._execution_plan = execution_plan
        self._main_script = self._prepare_script(execution_plan.Body)
        self._script_funcs = {}
        self._files_manager = fm.FilesManager(execution_plan)
        self._prepare_executors(execution_plan)

    def run(self):
        script_globals = {
            "args": bunch.Bunch(self._execution_plan.get('Parameters') or {})
        }
        script_globals.update(self._script_funcs)
        exec(self._main_script, script_globals)
        if '__execution_plan_exception' in script_globals:
            raise script_globals['__execution_plan_exception']
        return script_globals['__execution_plan_result']

    @staticmethod
    def _unindent(script, initial_indent):
        lines = script.expandtabs(4).split('\n')
        min_indent = sys.maxsize
        for line in lines:
            indent = -1
            for i, c in enumerate(line):
                if c != ' ':
                    indent = i
                    break
            if 0 <= indent < min_indent:
                min_indent = indent
        return '\n'.join([' ' * initial_indent + line[min_indent:]
                          for line in lines])

    def _prepare_executors(self, execution_plan):
        for key, value in execution_plan.Scripts.items():
            self._script_funcs[key] = script_runner.ScriptRunner(
                key, bunch.Bunch(value), self._files_manager)

    @staticmethod
    def _prepare_script(body):
        script = 'def __execution_plan_main():\n'
        script += ExecutionPlanRunner._unindent(body, 4)
        script += """
try:
    __execution_plan_result = __execution_plan_main()
except Exception as e:
    __execution_plan_exception = e
"""
        return script

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._files_manager.clear()
        return False



# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Murano-Agent
"""

from oslo_config import cfg
from oslo_log import log as logging

from muranoagent import version

CONF = cfg.CONF

opts = [
    cfg.StrOpt('storage',
               default='/var/murano/plans',
               help='Directory to store execution plans'),

    cfg.StrOpt('engine_key',
               help='Public key of murano-engine')
]

message_routing_opt = cfg.BoolOpt(
    'enable_dynamic_result_queue',
    help='Enable taking dynamic result queue from task field reply_to',
    default=False)


rabbit_opts = [
    cfg.HostAddressOpt('host',
                       help='The RabbitMQ broker address which used for '
                            'communication with Murano guest agents.',
                       default='localhost'),
    cfg.IntOpt('port', help='The RabbitMQ broker port.', default=5672),
    cfg.StrOpt('login',
               help='The RabbitMQ login.',
               default='guest'),
    cfg.StrOpt('password',
               help='The RabbitMQ password.',
               secret=True,
               default='guest'),
    cfg.StrOpt('virtual_host',
               help='The RabbitMQ virtual host.',
               default='/'),
    cfg.BoolOpt('ssl',
                help='Boolean flag to enable SSL communication through the '
                'RabbitMQ broker between murano-engine and guest agents.',
                default=False),
    cfg.StrOpt('ssl_version',
               default='',
               help='SSL version to use (valid only if SSL enabled). '
                    'Valid values are TLSv1 and SSLv23. SSLv2, SSLv3, '
                    'TLSv1_1, and TLSv1_2 may be available on some '
                    'distributions.'),
    cfg.StrOpt('ca_certs',
               help='SSL cert file (valid only if SSL enabled).',
               default=''),
    cfg.BoolOpt('insecure', default=False,
                help='This option explicitly allows Murano to perform '
                     '"insecure" SSL connections to RabbitMQ'),
    cfg.StrOpt('result_routing_key',
               help='This value should be obtained from API'),
    cfg.StrOpt('result_exchange',
               help='This value must be obtained from API',
               default=''),
    cfg.StrOpt('input_queue',
               help='This value must be obtained from API',
               default='')

]

CONF.register_opts(opts)
CONF.register_cli_opt(message_routing_opt)
CONF.register_opts(rabbit_opts, group='rabbitmq')
logging.register_options(CONF)


def parse_args(args=None, usage=None, default_config_files=None):
    version_string = version.version_info.version_string()
    CONF(args=args,
         project='muranoagent',
         version=version_string,
         usage=usage,
         default_config_files=default_config_files)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from muranoagent import executors as exe


class FunctionRunner(object):
    def __init__(self, name, script_runner):
        self._name = name
        self._script_executor = script_runner

    def __call__(self, *args, **kwargs):
        return self._script_executor.execute_function(
            self._name, *args, **kwargs)


class ScriptRunner(object):
    def __init__(self, name, script_info, files_manager):
        self._name = name
        self._executor = self._get_executor(script_info['Type'], name,
                                            script_info['EntryPoint'])
        self._script_info = script_info
        self._script_loaded = False
        self._files_manager = files_manager

    def __call__(self, *args, **kwargs):
        return self.execute_function(None, *args, **kwargs)

    def _get_executor(self, script_type, script_name, entry_point):
        create_executor = exe.Executors.create_executor
        if script_type != 'Application':
            executor = create_executor(script_type, entry_point)
        else:
            executor = create_executor(script_type, script_name)
        if executor is None:
            raise ValueError('The application type in {0} is not a valid '
                             'executor {1}'.format(script_name, script_type))
        return executor

    def execute_function(self, name, *args, **kwargs):
        self._load()
        return self._executor.run(name, *args, **kwargs)

    def __getattr__(self, item):
        return FunctionRunner(item, self)

    def _load(self):
        if not self._script_loaded:
            self._executor.load(
                self._prepare_files(),
                self._script_info.get("Options") or {})
            self._script_loaded = True

    def _prepare_files(self):
        for file_id in self._script_info.get('Files', []):
            self._files_manager.put_file(file_id, self._name)

        if self._script_info["Type"] == 'Application':
            return self._files_manager.put_file(
                self._script_info["EntryPoint"], self._name)
        return os.path.join(self._files_manager._cache_folder, self._name)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AgentException(Exception):
    def __init__(self, code, message=None, additional_data=None):
        self._error_code = code
        self._additional_data = additional_data
        super(AgentException, self).__init__(message)

    @property
    def error_code(self):
        return self._error_code

    @property
    def additional_data(self):
        return self._additional_data


class CustomException(AgentException):
    def __init__(self, code, message=None, additional_data=None):
        super(CustomException, self).__init__(
            code + 100, message, additional_data)


class IncorrectFormat(AgentException):
    pass



#    Copyright (c) 2015 Telefonica I+D
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import json
import os
from unittest import mock
from unittest.mock import ANY

from muranoagent import bunch
from muranoagent import exceptions as ex
from muranoagent.executors import chef
from muranoagent.tests.unit import base
from muranoagent.tests.unit import execution_plan as ep


class TestChefExecutor(base.MuranoAgentTestCase, fixtures.TestWithFixtures):
    def setUp(self):
        super(TestChefExecutor, self).setUp()
        self.chef_executor = chef.ChefExecutor('cookbook::recipe')

    def test_create_nodejson_noatts(self):
        """It tests the manifest without attributes."""
        node = self.chef_executor._create_manifest('cookbook', 'recipe', None)
        self.assertEqual(json.loads(node), self.get_node_no_atts())

    def test_create_nodejson(self):
        """It tests a manifest with attributes."""
        atts = {
            'att1': 'value1',
            'att2': 'value2'
        }

        node = self.chef_executor._create_manifest('cookbook', 'recipe', atts)
        self.assertEqual(json.loads(node), self.get_node_atts())

    @mock.patch('subprocess.Popen')
    @mock.patch('builtins.open')
    @mock.patch('os.path.exists')
    @mock.patch('os.path.isdir')
    def test_cookbook(self, mock_isdir, mock_exist, open_mock,
                      mock_subproc_popen):
        """It tests chef executor."""
        self._open_mock(open_mock)
        mock_exist.return_value = True
        mock_isdir.return_value = True

        process_mock = mock.Mock()
        attrs = {'communicate.return_value': ('output', 'ok'),
                 'poll.return_value': 0}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        template = self.useFixture(ep.ExPlanDownloable()).execution_plan
        script = list(template['Scripts'].values())
        self.chef_executor.load('path',
                                script[0]['Options'])
        self.chef_executor.run('test')

    @mock.patch('subprocess.Popen')
    @mock.patch('builtins.open')
    @mock.patch('os.path.exists')
    @mock.patch('os.path.isdir')
    def test_cookbook_error(self, mock_isdir, mock_exist, open_mock,
                            mock_subproc_popen):
        """It tests chef executor with error in the request."""
        self._open_mock(open_mock)
        mock_exist.return_value = True
        mock_isdir.return_value = True

        process_mock = mock.Mock()
        attrs = {'communicate.return_value': ('output', 'error'),
                 'poll.return_value': 2}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        template = self.useFixture(ep.ExPlanDownloable()).execution_plan
        script = list(template['Scripts'].values())[0]
        self.chef_executor.load('path',
                                script['Options'])
        self.assertRaises(ex.CustomException, self.chef_executor.run,
                          'test')

    def test_chef_cookbook_wrong(self):
        """It tests a wrong cookbook name."""
        chef_executor = chef.ChefExecutor('wrong')
        self.assertRaises(ex.CustomException, chef_executor.run,
                          'test')

    def test_chef_no_berkshelf(self):
        """It tests the cookbook path if Berkshelf is not enabled"""
        template = self.useFixture(ep.ExPlanDownloable()).execution_plan
        script = list(template['Scripts'].values())[0]
        self.chef_executor.load('path',
                                script['Options'])
        cookbook_path = self.chef_executor._create_cookbook_path('cookbook')
        self.assertEqual(cookbook_path, os.path.abspath('path'))

    @mock.patch('subprocess.Popen')
    @mock.patch('os.path.isfile')
    def test_chef_berkshelf_default_berksfile(self, mock_isfile,
                                              mock_subproc_popen):
        """It tests Berkshelf usage if no Berksfile path is provided"""
        mock_isfile.return_value = True

        process_mock = mock.Mock()
        attrs = {'communicate.return_value': ('output', 'ok'),
                 'poll.return_value': 0}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        template = self.useFixture(ep.ExPlanBerkshelf()).execution_plan
        script = list(template['Scripts'].values())[0]
        self.chef_executor.load('path',
                                script['Options'])
        self.chef_executor.module_name = 'test'
        cookbook_path = self.chef_executor._create_cookbook_path('cookbook')

        self.assertEqual(cookbook_path,
                         os.path.abspath('path/berks-cookbooks'))
        expected_command = 'berks vendor --berksfile={0} {1}'.format(
                           os.path.abspath('path/cookbook/Berksfile'),
                           cookbook_path)
        mock_subproc_popen.assert_called_once_with(expected_command,
                                                   cwd=ANY,
                                                   shell=ANY,
                                                   stdout=ANY,
                                                   stderr=ANY,
                                                   universal_newlines=ANY)

    @mock.patch('subprocess.Popen')
    @mock.patch('os.path.isfile')
    def test_chef_berkshelf_custom_berksfile(self, mock_isfile,
                                             mock_subproc_popen):
        """It tests Berkshelf usage if a custom Berksfile is provided"""
        mock_isfile.return_value = True

        process_mock = mock.Mock()
        attrs = {'communicate.return_value': ('output', 'ok'),
                 'poll.return_value': 0}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        template = self.useFixture(ep.ExPlanCustomBerskfile()).execution_plan
        script = list(template['Scripts'].values())[0]
        self.chef_executor.load('path',
                                script['Options'])
        self.chef_executor.module_name = 'test'
        cookbook_path = self.chef_executor._create_cookbook_path('cookbook')

        self.assertEqual(cookbook_path,
                         os.path.abspath('path/berks-cookbooks'))
        expected_command = 'berks vendor --berksfile={0} {1}'.format(
                           os.path.abspath('path/custom/customFile'),
                           cookbook_path)
        mock_subproc_popen.assert_called_once_with(expected_command,
                                                   cwd=ANY,
                                                   shell=ANY,
                                                   stdout=ANY,
                                                   stderr=ANY,
                                                   universal_newlines=ANY)

    @mock.patch('subprocess.Popen')
    @mock.patch('os.path.isfile')
    def test_chef_berkshelf_error(self, mock_isfile,
                                  mock_subproc_popen):
        """It tests if Berkshelf throws an error"""
        mock_isfile.return_value = True

        process_mock = mock.Mock()
        attrs = {'communicate.return_value': ('output', 'error'),
                 'poll.return_value': 2}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        template = self.useFixture(ep.ExPlanBerkshelf()).execution_plan
        script = list(template['Scripts'].values())[0]
        self.chef_executor.load('path',
                                script['Options'])
        self.chef_executor.module_name = 'test'
        self.assertRaises(ex.CustomException,
                          self.chef_executor._create_cookbook_path,
                          'cookbook')

    def _open_mock(self, open_mock):
        context_manager_mock = mock.Mock()
        open_mock.return_value = context_manager_mock
        file_mock = mock.Mock()
        file_mock.read.return_value = ''
        enter_mock = mock.Mock()
        enter_mock.return_value = file_mock
        exit_mock = mock.Mock()
        setattr(context_manager_mock, '__enter__', enter_mock)
        setattr(context_manager_mock, '__exit__', exit_mock)

    def _stub_uuid(self, values=None):
        class FakeUUID(object):
            def __init__(self, v):
                self.hex = v

        if values is None:
            values = []
        mock_uuid4 = mock.patch('uuid.uuid4').start()
        mock_uuid4.side_effect = [FakeUUID(v) for v in values]
        return mock_uuid4

    def get_template_downloable(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'file': {
                    'Name': 'myfile',
                    'URL': 'https://github.com'
                           '/apache/tomcat/blob/trunk/LICENSE',
                    'Type': 'Downloadable'
                }
            }
        )

    def get_template_file(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'test': {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'installer'
                }
            }
        )

    def get_node_atts(self):
        return {
            "run_list": [
                "recipe[cookbook::recipe]"
            ],
            "cookbook": {
                "att1": "value1",
                "att2": "value2"
            }
        }

    def get_node_no_atts(self):
        return {
            "run_list": [
                "recipe[cookbook::recipe]"
            ]
        }



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import timeutils
from oslo_utils import uuidutils

from muranoagent import exceptions as exc


class ExecutionResult(object):
    @staticmethod
    def from_result(result, execution_plan):
        if 'ID' not in execution_plan:
            raise ValueError('ID attribute is missing from execution plan')

        return {
            'FormatVersion': '2.0.0',
            'ID': uuidutils.generate_uuid(dashed=False),
            'SourceID': execution_plan.ID,
            'Action': 'Execution:Result',
            'ErrorCode': 0,
            'Body': result,
            'Time': str(timeutils.utcnow())
        }

    @staticmethod
    def from_error(error, execution_plan):
        if 'ID' not in execution_plan:
            raise ValueError('ID attribute is missing from execution plan')

        error_code = 1
        additional_info = None
        message = None
        if isinstance(error, int):
            error_code = error
        elif isinstance(error, Exception):
            message = str(error)
            if isinstance(error, exc.AgentException):
                error_code = error.error_code
                additional_info = error.additional_data

        return {
            'FormatVersion': '2.0.0',
            'ID': uuidutils.generate_uuid(dashed=False),
            'SourceID': execution_plan.ID,
            'Action': 'Execution:Result',
            'ErrorCode': error_code,
            'Body': {
                'Message': message,
                'AdditionalInfo': additional_info
            },
            'Time': str(timeutils.utcnow())
        }



#    Copyright (c) 2015 Telefonica I+D
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import git
from unittest import mock

from muranoagent import bunch
from muranoagent.common import config as cfg
from muranoagent import files_manager as fmanager
from muranoagent import script_runner
from muranoagent.tests.unit import base

CONF = cfg.CONF


class TestScriptRunner(base.MuranoAgentTestCase):
    def setUp(self):
        super(TestScriptRunner, self).setUp()
        CONF.set_override('storage', 'ss')

    @mock.patch('os.path.join')
    @mock.patch("muranoagent.files_manager.FilesManager")
    @mock.patch("muranoagent.executors.Executors")
    def test_script_runner_downloable(self, mock_file_manager, mock_executors,
                                      mock_os):
        mock_file_manager.put_file.return_value = None
        mock_executors.create_executor.return_value = None
        mock_os.return_value = '/tmp/1234'
        template = self.get_template_downloable_git()
        scripts = script_runner\
            .ScriptRunner('deploy',
                          template.get('Scripts')['deploy'],
                          mock_file_manager)
        scripts._prepare_files()

    def _stub_uuid(self, values=None):
        class FakeUUID(object):
            def __init__(self, v):
                self.hex = v

        if values is None:
            values = []
        mock_uuid4 = mock.patch('uuid.uuid4').start()
        mock_uuid4.side_effect = [FakeUUID(v) for v in values]
        return mock_uuid4

    def get_template_downloable_git(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'mycoockbook': {
                    'Name': 'mycoockbook.txt',
                    'URL': 'https://github.com/tomcat.git',
                    'Type': 'Downloadable'
                }
            },
            Scripts={
                'deploy': {
                    'EntryPoint': 'cookbook/recipe',
                    'Files': [
                        'https://github.com/tomcat.git',
                        {'java': 'https://github.com/java.git'}
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Chef',
                    'Version': '1.0.0'
                }
            }
        )

    def get_template_downloable(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'file': {
                    'Name': 'myfile',
                    'URL': 'https://github.com'
                           '/apache/tomcat/blob/trunk/LICENSE',
                    'Type': 'Downloadable'
                }
            }
        )

    def get_template_file(self):
        return bunch.Bunch(
            ID='ID',
            Files={
                'test': {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'installer'
                }
            }
        )



# Copyright (c) 2017 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import semantic_version

from muranoagent import exceptions as exc


max_format_version = semantic_version.Spec('<=2.2.0')


def validate_plan(plan):
    plan_format_version = semantic_version.Version(
        plan.get('FormatVersion', '1.0.0'))

    if plan_format_version not in max_format_version:
        # NOTE(kazitsev) this is Version in Spec not str in str
        raise exc.IncorrectFormat(
            9,
            "Unsupported format version {0} "
            "(I support versions {1})".format(
                plan_format_version, max_format_version))

    for attr in ('Scripts', 'Files'):
        if attr not in plan:
            raise exc.IncorrectFormat(
                2, '{0} is not in the execution plan'.format(attr))

    for attr in ('Scripts', 'Files', 'Options'):
        if attr in plan and not isinstance(
                plan[attr], dict):
            raise exc.IncorrectFormat(
                2, '{0} is not a dictionary'.format(attr))

    for name, script in plan.get('Scripts', {}).items():
        _validate_script(name, script, plan_format_version, plan)

    for key, plan_file in plan.get('Files', {}).items():
        _validate_file(plan_file, key, plan_format_version)


def _validate_script(name, script, plan_format_version, plan):
    for attr in ('Type', 'EntryPoint'):
        if attr not in script or not isinstance(script[attr], str):
            raise exc.IncorrectFormat(
                2, 'Incorrect {0} entry in script {1}'.format(
                    attr, name))

    if plan_format_version in semantic_version.Spec('>=2.0.0,<2.1.0'):
        if script['Type'] != 'Application':
            raise exc.IncorrectFormat(
                2, 'Type {0} is not valid for format {1}'.format(
                    script['Type'], plan_format_version))
        if script['EntryPoint'] not in plan.get('Files', {}):
            raise exc.IncorrectFormat(
                2, 'Script {0} misses entry point {1}'.format(
                    name, script['EntryPoint']))

    if plan_format_version in semantic_version.Spec('>=2.1.0'):
        if script['Type'] not in ('Application', 'Chef', 'Puppet'):
            raise exc.IncorrectFormat(
                2, 'Script has not a valid type {0}'.format(
                    script['Type']))
        if (script['Type'] == 'Application' and script['EntryPoint']
                not in plan.get('Files', {})):
            raise exc.IncorrectFormat(
                2, 'Script {0} misses entry point {1}'.format(
                    name, script['EntryPoint']))
        elif (script['Type'] != 'Application' and
              "::" not in script['EntryPoint']):
            raise exc.IncorrectFormat(
                2, 'Wrong EntryPoint {0} for Puppet/Chef '
                   'executors. :: needed'.format(script['EntryPoint']))

        for option in script['Options']:
            if option in ('useBerkshelf', 'berksfilePath'):
                if plan_format_version in semantic_version.Spec('<2.2.0'):
                    raise exc.IncorrectFormat(
                        2, 'Script has an option {0} invalid '
                           'for version {1}'.format(option,
                                                    plan_format_version))
                elif script['Type'] != 'Chef':
                    raise exc.IncorrectFormat(
                        2, 'Script has an option {0} invalid '
                           'for type {1}'.format(option, script['Type']))

    for additional_file in script.get('Files', []):
        mns_error = ('Script {0} misses file {1}'.format(
            name, additional_file))
        if isinstance(additional_file, dict):
            if (list(additional_file.keys())[0] not in
                    plan.get('Files', {}).keys()):
                raise exc.IncorrectFormat(2, mns_error)
        elif additional_file not in plan.get('Files', {}):
            raise exc.IncorrectFormat(2, mns_error)


def _validate_file(plan_file, key, format_version):
    if format_version in semantic_version.Spec('>=2.0.0,<2.1.0'):
        for plan in plan_file.keys():
            if plan in ('Type', 'URL'):
                raise exc.IncorrectFormat(
                    2, 'Download file is {0} not valid for this '
                       'version {1}'.format(key, format_version))

    if 'Type' in plan_file:
        for attr in ('Type', 'URL', 'Name'):
            if attr not in plan_file:
                raise exc.IncorrectFormat(
                    2,
                    'Incorrect {0} entry in file {1}'.format(attr, key))

    elif 'Body' in plan_file:
        for attr in ('BodyType', 'Body', 'Name'):
            if attr not in plan_file:
                raise exc.IncorrectFormat(
                    2, 'Incorrect {0} entry in file {1}'.format(
                        attr, key))

        if plan_file['BodyType'] not in ('Text', 'Base64'):
            raise exc.IncorrectFormat(
                2, 'Incorrect BodyType in file {0}'.format(key))
    else:
        raise exc.IncorrectFormat(
            2, 'Invalid file {0}: {1}'.format(
                key, plan_file))



#    Copyright (c) 2015 Telefonica I+D
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from muranoagent.tests.unit import base
from muranoagent import util


class TestUtils(base.MuranoAgentTestCase):

    def test_str_to_bytes(self):
        self.assertEqual(util._to_bytes('test'), b'test')

    def test_bytes_to_bytes(self):
        self.assertEqual(util._to_bytes(b'test'), b'test')

    def test_b64encode_str(self):
        self.assertEqual(util.b64encode('test'), 'dGVzdA==')

    def test_b64encode_bytes(self):
        self.assertEqual(util.b64encode(b'test'), 'dGVzdA==')

    def test_b64decode_str(self):
        self.assertEqual(util.b64decode('dGVzdA=='), 'test')

    def test_b64decode_bytes(self):
        self.assertEqual(util.b64decode(b'dGVzdA=='), 'test')



# Copyright (c) 2015 Telefonica I+D.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fixtures

from muranoagent import bunch


class ExPlanDownloable(fixtures.Fixture):
    def setUp(self):
        super(ExPlanDownloable, self).setUp()
        self.execution_plan = bunch.Bunch(
            Action='Execute',
            Body='return deploy(args.appName).stdout\n',
            Files={
                'ID1': {
                    'Name': 'tomcat.git',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/tomcat.git'
                },
                'ID2': {
                    'Name': 'java',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/java.git'
                },

            },
            FormatVersion='2.1.0',
            ID='ID',
            Name='Deploy Chef',
            Parameters={
                'appName': '$appName'
            },
            Scripts={
                'deploy': {
                    'EntryPoint': 'cookbook::recipe',
                    'Files': [
                        'ID1',
                        'ID2'
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Chef',
                    'Version': '1.0.0'
                }
            },
            Version='1.0.0'
        )
        self.addCleanup(delattr, self, 'execution_plan')


class ExPlanApplication(fixtures.Fixture):
    def setUp(self):
        super(ExPlanApplication, self).setUp()
        self.execution_plan = bunch.Bunch(
            Action='Execute',
            Body='return deploy(args.appName).stdout',
            Files={
                'ID1': {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTomcat.sh'
                },
                'ID2': {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'installer'
                },
                'ID3': {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'common.sh'
                }
            },
            FormatVersion='2.1.0',
            ID='ID',
            Name='Deploy Tomcat',
            Parameters={
                'appName': '$appName'
            },
            Scripts={
                'deploy': {
                    'EntryPoint': 'ID1',
                    'Files': [
                        'ID2',
                        'ID3'
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            Version='1.0.0'
        )
        self.addCleanup(delattr, self, 'execution_plan')


class ExPlanDownloableWrongFormat(fixtures.Fixture):
    def setUp(self):
        super(ExPlanDownloableWrongFormat, self).setUp()
        self.execution_plan = bunch.Bunch(
            ID='ID',
            FormatVersion='0.0.0'
        )
        self.addCleanup(delattr, self, 'execution_plan')


class ExPlanDownloableNoFiles(fixtures.Fixture):
    def setUp(self):
        super(ExPlanDownloableNoFiles, self).setUp()
        self.execution_plan = bunch.Bunch(
            ID='ID',
            FormatVersion='2.1.0',
            Scripts={
                'deploy': {
                    'EntryPoint': 'cookbook::recipe',
                    'Files': [
                        'https://github.com/tomcat.git',
                        {'java': 'https://github.com/java.git'}
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Chef',
                    'Version': '1.0.0'
                }
            }
        )
        self.addCleanup(delattr, self, 'execution_plan')


class PuppetExPlanDownloable(fixtures.Fixture):
    def setUp(self):
        super(PuppetExPlanDownloable, self).setUp()
        self.execution_plan = bunch.Bunch(
            Action='Execute',
            Body='return deploy(args.appName).stdout\n',
            Files={
                'ID1': {
                    'Name': 'tomcat.git',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/tomcat.git'
                },
                'ID2': {
                    'Name': 'java',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/java.git'
                },

            },
            FormatVersion='2.0.0',
            ID='ID',
            Name='Deploy Puppet',
            Parameters={
                'appName': '$appName'
            },
            Scripts={
                'deploy': {
                    'EntryPoint': 'cookbook::recipe',
                    'Files': [
                        'ID1',
                        'ID2'
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Puppet',
                    'Version': '1.0.0'
                }
            },
            Version='1.0.0'
        )
        self.addCleanup(delattr, self, 'execution_plan')


class ExPlanBerkshelf(fixtures.Fixture):
    def setUp(self):
        super(ExPlanBerkshelf, self).setUp()
        self.execution_plan = bunch.Bunch(
            Action='Execute',
            Body='return deploy(args.appName).stdout\n',
            Files={
                'ID1': {
                    'Name': 'tomcat.git',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/tomcat.git'
                }
            },
            FormatVersion='2.2.0',
            ID='ID',
            Name='Deploy Chef',
            Parameters={},
            Scripts={
                'deploy': {
                    'EntryPoint': 'cookbook::recipe',
                    'Files': [
                        'ID1'
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True,
                        'useBerkshelf': True
                    },
                    'Type': 'Chef',
                    'Version': '1.0.0'
                }
            },
            Version='1.0.0'
        )
        self.addCleanup(delattr, self, 'execution_plan')


class ExPlanCustomBerskfile(fixtures.Fixture):
    def setUp(self):
        super(ExPlanCustomBerskfile, self).setUp()
        self.execution_plan = bunch.Bunch(
            Action='Execute',
            Body='return deploy(args.appName).stdout\n',
            Files={
                'ID1': {
                    'Name': 'tomcat.git',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/tomcat.git'
                }
            },
            FormatVersion='2.2.0',
            ID='ID',
            Name='Deploy Chef',
            Parameters={},
            Scripts={
                'deploy': {
                    'EntryPoint': 'cookbook::recipe',
                    'Files': [
                        'ID1'
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True,
                        'useBerkshelf': True,
                        'berksfilePath': 'custom/customFile'
                    },
                    'Type': 'Chef',
                    'Version': '1.0.0'
                }
            },
            Version='1.0.0'
        )
        self.addCleanup(delattr, self, 'execution_plan')


class ExPlanBerkWrongVersion(fixtures.Fixture):
    def setUp(self):
        super(ExPlanBerkWrongVersion, self).setUp()
        self.execution_plan = bunch.Bunch(
            Action='Execute',
            Body='return deploy(args.appName).stdout\n',
            Files={
                'ID1': {
                    'Name': 'tomcat.git',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/tomcat.git'
                }
            },
            FormatVersion='2.1.0',
            ID='ID',
            Name='Deploy Chef',
            Parameters={},
            Scripts={
                'deploy': {
                    'EntryPoint': 'cookbook::recipe',
                    'Files': [
                        'ID1'
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True,
                        'useBerkshelf': True
                    },
                    'Type': 'Chef',
                    'Version': '1.0.0'
                }
            },
            Version='1.0.0'
        )
        self.addCleanup(delattr, self, 'execution_plan')
